require 'circle'
require 'rect'

local BRANCH_SIZE = 20
local SEARCH_DISTANCE = 100
local HUNGER_RATE = 2
local MAX_CREATURES = 100

function Branch(creature, parent, length, angle)
	local branch = {
		x = parent.x,
		y = parent.y,
		parent = parent,
		endpoint = nil,
		length = length,
		angle = angle
	}

	function branch:click(mx, my, button)
		return self.endpoint:click(mx, my, button)
	end

	function branch:draw()
		local r, g, b, a = love.graphics.getColor()

		local ex = math.sin(math.rad(self.angle))*self.length
		local ey = math.cos(math.rad(self.angle))*self.length

		love.graphics.setColor({r, g, b, a/2})
		love.graphics.setLineWidth(10)

		love.graphics.line(self.x, self.y, self.endpoint.x, self.endpoint.y)

		love.graphics.setColor({r, g, b, a})
		love.graphics.setLineWidth(3)
		love.graphics.line(self.x, self.y, self.endpoint.x, self.endpoint.y)

		self.endpoint:draw()
	end

	function branch:placeEndpoint()
		self.endpoint.x = self.x + math.sin(math.rad(self.angle)) * self.length
		self.endpoint.y = self.y + math.cos(math.rad(self.angle)) * self.length
	end

	function branch:update(dtime)
		self.x = self.parent.x
		self.y = self.parent.y

		self:placeEndpoint()

		local turn = 200
		self.angle = self.angle+(math.random()*turn-turn/2) * dtime

		self.endpoint:update(dtime)
	end

	branch.endpoint = Joint(creature, branch)
	branch:placeEndpoint()
	return branch
end

function Joint(creature, parent)
	local x = 0
	local y = 0
	local size = 6

	local joint = {
		x = x,
		y = y,
		parent = parent,
		nodes = {},
		size = size,
		visible = false,
		leaf = true,
		circle = Circle(x, y, size),
		deathTimer = 0,
		creature = creature,
		alpha = 0xff,
		maxAge = 0.5
	}

	function joint:addBranch(length, angle)
		local branch = Branch(self.creature, self, length, angle)
		self.visible = true
		self.leaf = false
		table.insert(self.nodes, branch)
		return branch
	end

	function joint:click(mx, my, button)
		local dx = math.abs(self.x - mx)
		local dy = math.abs(self.y - my)

		if dx*dx + dy*dy < self.size*self.size then
			return true
		end
		for _,n in ipairs(self.nodes) do
			if n:click(mx, my, button) then return true end
		end
		return false
	end

	function joint:destroy()
		self.parent.parent:remove(self.parent)
	end

	function joint:draw()
		if not self.visible then return end

		local colour = self.creature.colour
		if self.destroyed then
			colour = {colour[1], colour[2], colour[3], self.alpha}
		end

		love.graphics.setColor({colour[1], colour[2], colour[3], colour[4]/2})
		love.graphics.setLineWidth(10)
		love.graphics.circle('line', self.x, self.y, self.size)

		love.graphics.setColor(colour)
		love.graphics.setLineWidth(3)
		love.graphics.circle('line', self.x, self.y, self.size)

		for _,n in ipairs(self.nodes) do
			n:draw()
		end
	end

	function joint:mark()
		self.destroyed = true
		for _,n in ipairs(self.nodes) do
			n.endpoint:mark()
		end
	end

	function joint:remove(node)
		for i=1,#self.nodes do
			if self.nodes[i] == node then
				table.remove(self.nodes, i)
				return
			end
		end
	end

	function joint:update(dtime)
		for _,n in ipairs(self.nodes) do
			n:update(dtime)
		end
		self.circle.x = self.x
		self.circle.y = self.y

		if self.destroyed then
			self.deathTimer = self.deathTimer+dtime
			self.alpha = math.min((1 - (self.deathTimer / self.maxAge)) * 0xff, 0xff)
			if self.deathTimer >= self.maxAge then
				self.removed = true
			end
		end
	end

	return joint
end

function Creature(x, y, genes, food, angle, hostile, colour)
	local creature = {
		x = x,
		y = y,
		genotype = genes,
		root = nil,
		colour = colour or {0x20, 0xff, 0x40, 0xff},
		angle = angle or math.random()*360,
		dx = 0,
		dy = 0,
		active = false,
		joints = {},
		food = food or 255,
		searchDistance = SEARCH_DISTANCE,
		canSpawn = true,
		edgeBounce = true,
		hostile = hostile,
		tag = 'creature',
		deathTimer = 0
	}
	creature.root = Joint(creature)
	if not creature.hostile then
		music.setSequence(genes)
	end

	function creature:addJoint(j)
		table.insert(self.joints, j)
	end

	function creature:build()
		self.root.x = self.x
		self.root.y = self.y

		local size = 1

		local joint = self.root

		self:addJoint(self.root)

		for i = 1, #self.genotype do
			local c = self.genotype:sub(i,i)

			if c == '1' then
				joint = joint:addBranch(BRANCH_SIZE, 45).endpoint
				self:addJoint(joint)
				size = size+1
			elseif c == '2' then
				joint = joint:addBranch(BRANCH_SIZE, 180-45).endpoint
				self:addJoint(joint)
				size = size+1
			elseif c == '3' then
				local j = joint:addBranch(BRANCH_SIZE, 45).endpoint
				self:addJoint(j)

				joint = joint:addBranch(BRANCH_SIZE, -45).endpoint
				self:addJoint(joint)

				size = size+2
			elseif c == '4' then
				joint = self.root
			end
		end

		self.size = size
		local flagella = 0
		for _,j in ipairs(self.joints) do
			if j.leaf then flagella = flagella+1 end
		end
		self.speed = math.max(flagella * 10 - size*2, 0)
		self.searchDistance = 20 * self.size
	end

	function creature:checkIntersection(c)
		for _,j in ipairs(self.joints) do
			for _,cj in ipairs(c.joints) do
				if not j.destroyed and not c.destroyed and j.circle:intersects(cj.circle) then
					cj:mark()
					j:mark()
				end
			end
		end
		c:removeJoints()
		self:removeJoints()
	end

	function creature:clone()
		if not self.canSpawn then return end
		if #scene:getTag('creature') > MAX_CREATURES then return end
		scene:add(Creature(self.x, self.y, genotype.mutate(self.genotype), self.food / 2, self.angle, self.hostile))
		self.food = self.food / 2
	end

	function creature:destroy()
		self.destroyed = true
	end

	function creature:draw()
		love.graphics.setBlendMode('additive')
		love.graphics.setColor(self.colour)
		self.root:draw()
	end

	function creature:rect()
		local minX = self.x
		local maxX = self.x
		local minY = self.y
		local maxY = self.y

		for _,j in ipairs(self.joints) do
			if j.x < minX then minX = j.x - j.size end
			if j.x > maxX then maxX = j.x + j.size end
			if j.y < minY then minY = j.y - j.size end
			if j.y > maxY then maxY = j.y + j.size end
		end
		return Rect(minX, minY, maxX - minX, maxY - minY)
	end

	function creature:removeJoints()
		local newJoints = {}
		for _,j in ipairs(self.joints) do
			if j.removed then
				if j.parent then
					j:destroy()
				else
					self:destroy()
				end
			else
				table.insert(newJoints, j)
			end
		end
		self.joints = newJoints
		self.x = self.root.x
		self.y = self.root.y
		if #self.joints == 0 then
			scene:remove(self)
		end
	end

	function creature:update(dtime)
		self.active = true
		self.dx = math.sin(math.rad(self.angle)) * self.speed
		self.dy = math.cos(math.rad(self.angle)) * self.speed
		if not self.destroyed then
			self.x = self.x+self.dx * dtime
			self.y = self.y+self.dy * dtime
		end
		self.root.x = self.x
		self.root.y = self.y
		self.root:update(dtime)

		local screen = scene:rect()

		local nearest = nil
		local nearestDistance = 0

		local rect = self:rect()
		if self.edgeBounce then
			if rect.x + rect.w >= screen.w then
				self.dx = -self.dx
				self.angle = math.deg(math.atan2(self.dx, self.dy))
				self.x = self.x-(rect.x+rect.w-screen.w)
			end

			if rect.x <= 0 then
				self.dx = -self.dx
				self.angle = math.deg(math.atan2(self.dx, self.dy))
				self.x = self.x-rect.x
			end

			if rect.y + rect.h >= screen.h then
				self.dy = -self.dy
				self.angle = math.deg(math.atan2(self.dx, self.dy))
				self.y = self.y-(rect.y+rect.h-screen.h)
			end

			if rect.y <= 0 then
				self.dy = -self.dy
				self.angle = math.deg(math.atan2(self.dx, self.dy))
				self.y = self.y-rect.y
			end
		end

		local foods = scene:getTag('food')
		local creatures = scene:getTag('creature')
		for _,joint in ipairs(self.joints) do
			if not joint.destroyed then
				for _,f in ipairs(foods) do
					local fdx = joint.x - f.x
					local fdy = joint.y - f.y
					local sqdistance = fdx*fdx+fdy*fdy

					if (not f.removed) and (nearest == nil or sqdistance < nearestDistance) then
						nearest = f
						nearestDistance = sqdistance
					end
				end

				local turnSpeed = self.speed

				if nearest then
					if joint.circle:intersects(nearest.circle) then
						scene:remove(nearest)
						nearest = nil
						nearestDistance = 0
						self.food = self.food + 100
					elseif nearestDistance < self.searchDistance*self.searchDistance then
						local angle = -math.deg(math.atan2(joint.x-nearest.x, -(joint.y-nearest.y)))

						self.angle = angle
						if self.angle > angle then
							self.angle = self.angle - turnSpeed * dtime
						elseif self.angle < angle then
							self.angle = self.angle + turnSpeed * dtime
						end
					end
				end
			end
		end

		if self.hostile then
			for _,c in ipairs(creatures) do
				if self:rect():intersects(c:rect()) then
					self:checkIntersection(c)
				end
			end
		end

		self.food = self.food - HUNGER_RATE * dtime

		if self.food > 0xff then
			self:clone()
		end

		if self.food <= 0 then
			scene:remove(self)
		else
			self.colour[4] = math.min((self.food / 0xff) * 0xff, 0xff)
		end

		self:removeJoints()
	end

	creature:build()
	return creature
end
